'''
Copyright (C) Espec North America, INC. - All Rights Reserved
Written by Myles Metzler mmetzler@espec.com, Feb. 2016

Partial modbus implimantation for communicating with watlow controllers (holding registers only)
'''
#pylint: disable=W0703
import socket
import struct
import time
import collections
import serial

class ModbusError(Exception):
    '''Generic Modbus exception.'''
    pass

class Modbus(object):
    '''
    A subset of a modbus master library, only impliments modbus functions:
    3: Read Holding Register(s)
    6: Write Holding Register
    16: Write Multiple Holding Registers
    '''
    errorMessages = {
        1: 'Illegal Function',
        2: 'Illegal Data Address',
        3: 'Illegal Data Value',
        4: 'Slave Device Failure',
        5: 'Acknowledge',
        6: 'Slave Device Busy',
        7: 'Negative Acknowledge',
        8: 'Memory Parity Error',
        10:'Gateway Path Unavalable',
        11:'Gateway Target Device Failed To Respond'
    }
    address = 1
    retry = False


    def read_holding(self, register, count=1):
        '''
        Read holding register(s)

        Args:
            register (int): The modbus register to read
            count (int): The number of modbus registers to read (default=1)

        Returns:
            list. unsigned 16bit integers
        '''
        packet = self.make_packet(3, register, count)
        try:
            rval = self.interact(packet)
        except ModbusError:
            if self.retry:
                rval = self.interact(packet)
            else:
                raise
        return self.decode_packet(rval, packet)

    def read_holding_signed(self, register, count=1):
        '''
        Read some signed short(s)

        Args:
            register (int): The modbus register to read
            count (int): The number of modbus registers to read (default=1)

        Returns:
            list. signed 16bit integers
        '''
        vals = self.read_holding(register, count)
        return [struct.unpack('h', struct.pack('H', val))[0] for val in vals]

    def read_holding_float(self, register, count=1):
        '''
        Read some floating point values from 2 adjacent modbus registers

        Args:
            register (int): the first register to start reading at.
            count (int): the number of floats to read (2*count will actually be read)

        Returns:
            list. 32bit floats
        '''
        val = self.read_holding(register, count*2)
        return [round(struct.unpack('f', struct.pack('HH', val[i], val[i+1]))[0], 1)
                for i in range(0, count*2, 2)]

    def read_holding_string(self, register, count):
        '''
        Read a string

        Args:
            register (int): The register to start reading from
            count(int): The number of registers to read (length of string)

        Returns:
            str
        '''
        val = self.read_holding(register, count)
        rstring = ""
        for char in val:
            if char is not 0:
                rstring = rstring + chr(char)
        return rstring

    def write_holding(self, register, value):
        '''
        Write to holding 16bit register(s), accepts single values or lists of values

        Args:
            register (int): register(s) to write to
            value (int or list(int)): value(s) to write,
        '''
        packettype = 16 if isinstance(value, collections.Iterable) else 6
        packet = self.make_packet(packettype, register, value)
        try:
            rval = self.interact(packet)
        except ModbusError:
            if self.retry:
                rval = self.interact(packet)
            else:
                raise
        self.decode_packet(rval, packet)

    def write_holding_signed(self, register, value):
        '''
        Write to signed 16bit holding register(s), accepts single values or lists of values

        Args:
            register (int): register(s) to write to
            value (int or list(int)): value(s) to write,
        '''
        if isinstance(value, collections.Iterable):
            value = [0xFFFF & val for val in value]
        else:
            value = 0xFFFF & value #trim to 16bit signed int
        self.write_holding(register, value)

    def write_holding_float(self, register, value):
        '''
        Write floating point values to the controller

        Args:
            register (int): first register to write to, 2 float value will be written.
            value (float or list(float)): vlaue(s) to write to
        '''
        if isinstance(value, collections.Iterable):
            packval = ''.join([struct.unpack('HH', struct.pack('f', val)) for val in value])
        else:
            packval = struct.unpack('HH', struct.pack('f', value))
        self.write_holding(register, packval)

    def write_holding_string(self, register, value, length=20, padder=0):
        '''
        Write a string to the controller

        Args:
            register (int): first register to wrote to
            value (str): The string to write
            length (int): The string will be padded or truncated to this length.

        Return:
            None
        '''
        mods = [ord(c) for c in value]
        mods.extend([padder]*length)
        self.write_holding(register, mods[0:length])

    def interact(self, packet):
        '''Interact with the physical interface'''
        raise NotImplementedError

    def make_packet(self, function, register, args):
        '''Make modbus request packet.'''
        if function == 3:
            return struct.pack(">BBHH", self.address, function, register, args)
        elif function == 6:
            return struct.pack(">BBHH", self.address, function, register, args)
        elif function == 16:
            data = struct.pack(">BBHHB", self.address, function, register, len(args), len(args)*2)
            for val in args:
                data += struct.pack(">H", val)
            return data
        else:
            raise NotImplementedError("Only modbus function codes 3,6,16 are implimented.")

    def decode_packet(self, packet, spacket):
        '''Decode the modbus request packet.'''
        fcode = struct.unpack(">B", packet[1])[0]
        addr = struct.unpack(">B", packet[0])[0]
        if self.address != addr:
            shex = ":".join("{:02x}".format(ord(c) for c in spacket))
            rhex = ":".join("{:02x}".format(ord(c) for c in packet))
            raise ModbusError("Address error; Sent=%s, Recieved=%s" % (shex, rhex))
        if fcode > 127:
            ecode = struct.unpack(">B", packet[2])[0]
            ttp = (ecode, self.errorMessages.get(ecode, 'Unknown error code'))
            raise ModbusError('Modbus Error: Exception code = %d(%s)' % ttp)

        if fcode == 3: #Read holding register(s)
            cnt = struct.unpack(">B", packet[2])[0]/2
            return struct.unpack(">%dH" % cnt, packet[3:])
        elif fcode == 6:
            pass #nothing is required
        elif fcode == 16:
            pass #nothing required
        else:
            raise NotImplementedError("Only modbus function codes 3,6,16 are implimented.")

class ModbusRTU(Modbus):
    '''
    A subset of a modbus RTU master library, only impliments modbus functions:
    3: Read Holding Register(s)
    6: Write Holding Register
    16: Write Multiple Holding Registers
    '''

    def __init__(self, address, port, **kwargs):
        self.address = address
        self.retry = kwargs.get('retry', True)
        #watlow suggests using 0.012 char send time for buads greater than 19200
        databits, stopbits = kwargs.get('databits', 8), kwargs.get('stopbits', 1)
        baud = kwargs.get('baud', 9600)
        # calculated pause time does not work on the Watlow F4T, using watlow recomended delay...
        #self.pause = 3.5 * (((databits + stopbits + 2)/ baud) if baud < 19200 else 0.012)
        self.pause = 0.03
        self.serial = serial.Serial(
            port=port,
            baudrate=baud,
            bytesize=databits,
            parity=kwargs.get('parity', 'N'),
            stopbits=stopbits,
            timeout=kwargs.get('timeout', 3)
        )

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def close(self):
        '''
        Close the serial port.
        '''
        self.serial.close()

    def __calc_crc(self, data):
        '''
        calculate the CRC16
        '''
        crc = 0xFFFF
        for i in data:
            crc = crc ^ ord(i)
            for _ in xrange(8):
                tmp = crc & 1
                crc = crc >> 1
                if tmp:
                    crc = crc ^ 0xA001
        return ((crc % 256) << 8) + (crc >> 8) #swap byte order

    def __check_crc(self, data, crc):
        '''
        check a given set of data against a given crc value
        '''
        crc = (crc[0] << 8) + crc[1]
        if crc != self.__calc_crc(data):
            raise ModbusError("The CRCs do not match.")

    def interact(self, packet):
        crc = struct.pack(">H", self.__calc_crc(packet))
        self.serial.write(packet + crc)
        time.sleep(self.pause)
        head = self.serial.read(2)
        if len(head) == 0:
            raise ModbusError("The slave device did not respond.")
        raddress = struct.unpack('>B', head[0])[0]
        fcode = struct.unpack('>B', head[1])[0]
        if fcode == 16 or fcode == 6:
            body = self.serial.read(4)
        elif fcode == 3:
            body = self.serial.read(1)
            body += self.serial.read(struct.unpack('>B', body)[0])
        elif fcode > 127:
            body = self.serial.read(1)
        else:
            raise NotImplementedError("Only modbus function codes 3,6,16 are implimented.")
        rcrc = struct.unpack('>H', self.serial.read(2))[0]
        ccrc = self.__calc_crc(head+body)
        if self.address != raddress:
            shex = ":".join(["{:02x}".format(ord(c)) for c in packet+crc])
            rhex = ":".join(["{:02x}".format(ord(c)) for c in head+body+rcrc])
            raise ModbusError("Address error; Sent=%s, Recieved=%s" % (shex, rhex))
        if rcrc != ccrc:
            shex = ":".join(["{:02x}".format(ord(c)) for c in packet+crc])
            rhex = ":".join(["{:02x}".format(ord(c)) for c in head+body+rcrc])
            raise ModbusError("CRC error; Sent=%s, Recieved=%s" % (shex, rhex))
        return head + body

class ModbusTCP(Modbus):
    '''
    A subset of a modbus TCP master library, only impliments modbus functions:
    3: Read Holding Register(s)
    6: Write Holding Register
    16: Write Multiple Holding Registers
    '''

    def __init__(self, address, host, port=502, **kwargs):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        #self.socket.settimeout(timeout)
        self.socket.setblocking(True)
        self.socket.connect((host, port))
        self.packet_id = 1
        self.address = address
        time.sleep(0.1)

    def __del__(self):
        self.close()

    def close(self):
        '''
        Close the tcp socket.
        '''
        self.socket.close()
        time.sleep(0.1)

    def __make_mbap(self, length):
        '''
        make the modbus mbap
        '''
        return struct.pack(">3H", self.packet_id, 0, length)

    def interact(self, packet):
        '''
        interact with the slave device
        '''
        self.socket.send(self.__make_mbap(len(packet)) + packet)
        mbap_raw = self.socket.recv(6)
        if len(mbap_raw) == 0:
            raise ModbusError("The controller did not respond to the request (MBAP length = 0)")
        if len(mbap_raw) != 6:
            ttp = (len(mbap_raw), mbap_raw)
            raise ModbusError("MBAP length error; expected:6, got:%s (%r)" % ttp)
        mbap = struct.unpack('>3H', mbap_raw)
        body = self.socket.recv(mbap[2])
        if mbap[0] != self.packet_id:
            ttp = (self.packet_id, mbap[0], mbap_raw)
            raise ModbusError("MBAP id error; expected:%r, got:%r (%r)" % ttp)
        #self.packet_id = self.packet_id + 1 if self.packet_id < 65535 else 0
        return body



'''
Upper level interface for the Watlow F4 controller

:copyright: (C) Espec North America, INC.
:license: MIT, see LICENSE for more details.
'''
#pylint: disable=W0703
import datetime
import re
import time
from chamberconnectlibrary.modbus import ModbusError, ModbusRTU, ModbusTCP
from chamberconnectlibrary.controllerinterface import ControllerInterface, exclusive
from chamberconnectlibrary.controllerinterface import ControllerInterfaceError

class WatlowF4(ControllerInterface):
    '''
    A class for interfacing with Watlow F4T

    Kwargs:
        interface (str): The connection method::
            "TCP" -- modbusTCP (default)
            "RTU" -- modbusRTU
        adr (int): The modbus address of the controller (default=1)
        host (str): The hostname (IP address) of the Watlow F4T when interface="TCP"
        serialport (str): The serial port to use when interface="RTU"
        baudrate (int): The serial port's baud rate to use when interface="RTU"
        loops (int): The number of control loops the controller has (default=1, max=4)
        cascades (int): The number of cascade control loops the controller has (default=0, max=3)
        cond_event (int): The event # used to the controller on/off (default=0(disabled))
        limits (list(bool)): list of slots where limit controllers are installed (default=[5])
        loop_event (list(int)): list of events #'s that enable/disble loops (default=[0,2,0,0])
        cascade_event (list(int)): list of events #'s that enbl/dis casc.lps(default=[0,0,0,0])
        cascade_ctl_event (list(int)): list of event#'s enbl/dis casc. mode (default=[0,0,0,0])
        time_zone (None): Not currently used
        alarms (int): The number of alarms that the controller has (default=6)
        profiles (bool): If True the controller supports profiles(programs) (default=False)
        lock (RLock): The locking method to use when accessing the controller (default=RLock())
    '''

    def __init__(self, **kwargs):
        self.iwatlow_val_dict, self.client, self.loops, self.cascades = None, None, None, None
        self.init_common(**kwargs)
        self.cond_event = kwargs.get('cond_event')
        self.limits = kwargs.get('limits', [])

        #list of events that may enable or disable a loop index 0=loop1, events=1-8 0=not used
        self.loop_event = kwargs.get('loop_event', [0, 0, 0, 0])

        #list of events that may enable or disable a cascade loop
        self.cascade_event = kwargs.get('cascade_event', [0, 0, 0, 0])

        self.combined_event = [
            self.cascade_event[0] if self.cascades > 0 else self.loop_event[0],
        ]
        if self.loops == 2:
            self.combined_event.append(self.loop_event[1])
        elif self.cascades == 1 and self.loops == 1:
            self.combined_event.append(self.loop_event[0])

        self.events = 8

        self.named_loop_map_list = kwargs.get('loop_names', [])

        #these are detectable from the part number (call process_partno())
        self.alarms = kwargs.get('alarms', 6)
        self.profiles = kwargs.get('profiles', False)
        self.scalar = [None, None, None]
        self.__update_loop_map()


    def __range_check(self, val, minimum, maximum):
        '''
        min <= val <= max

        Args:
            val (int): the value to check the range of
            min (int): the minimum possible value
            max (int): the maximum possible value
        Returns:
            None
        Raises:
            ValueError
        '''
        if maximum < minimum or val < minimum or val > maximum:
            raise ValueError("Index is not within bounds or bounds are not valid")

    def __update_loop_map(self):
        '''
        update the loop map.
        '''
        if self.cascades > 0:
            self.loop_map = [{'type':'cascade', 'num':1}]
        else:
            self.loop_map = [{'type':'loop', 'num':1}]
        if self.cascades + self.loops > 1:
            self.loop_map += [{'type':'loop', 'num':2}]
        if len(self.named_loop_map_list) == len(self.loop_map):
            self.named_loop_map = {name:i for i, name in enumerate(self.named_loop_map_list)}

    def __get_scalar(self, N):
        '''
        Get the scalar factor for a analog input

        Args:
            N (int): 1 or 2
        Returns:
            float
        '''
        try:
            if self.scalar[N-1] is None:
                my_scalar = self.client.read_holding([606, 616, 626][N-1])[0]*10
                self.scalar[N-1] = 1 / float(my_scalar if my_scalar > 0 else 1)
            return self.scalar[N-1]
        except Exception:
            return 1

    def __get_digital_input(self, N):
        '''
        Get the state of a digital input

        Args:
            N (int): 1-4
        Returns:
            Boolean
        '''
        tval = self.client.read_holding(1061 + 2*(N-1))[0]
        return self.client.read_holding(201 + 12*(N-1))[0] != tval

    def __set_prgm_name(self, N, value):
        '''write a profile name, does not save eeprom'''
        value = value.upper()
        value = value.rstrip()
        if len(value) > 10:
            value = value[0:10]
        if re.match("^[A-Z0-9 ]*$", value):
            self.client.write_holding_string(3500+10*(N-1), value, 10, 32)
        else:
            raise ValueError('Name must be uppercase letters and numbers only.')

    def __get_analog_input_setup(self):
        lookup = ['thermocouple', 'rtd', 'process', 'wetbulb-drybulb', 'off']
        try:
            inputs = [lookup[self.client.read_holding(600)[0]]]
            inputs.append(lookup[self.client.read_holding(610)[0]])
            inputs.append(lookup[self.client.read_holding(620)[0]])
        except ModbusError:
            pass
        return inputs

    def __edit_prgm_step_autostart(self, step):
        '''setup an autostart step'''
        daylookup = ['daily', 'sun', 'mon', 'tue', 'wed', 'thur', 'fri', 'sat']
        self.client.write_holding(4003, 0)
        if step['start_type'] == 'day':
            self.client.write_holding(4004, 1)
            self.client.write_holding(4008, daylookup.index(step['start_time']['day']))
        else:
            date = [
                0,
                step['start_date']['month'],
                step['start_date']['day'],
                step['start_date']['year']
            ]
            self.client.write_holding(4004, date)
        my_time = [
            step['start_time']['hours'],
            step['start_time']['minutes'],
            step['start_time']['seconds']
        ]
        self.client.write_holding(4009, my_time)

    def __edit_prgm_step_rampsoak(self, step):
        eventlookup = ['disable', 'off', 'on']
        stypes = {'ramptime':1, 'ramprate':2, 'soak':3}
        self.client.write_holding(4003, stypes[step['type']])
        if step['wait']['enable']:
            self.client.write_holding(4012, 1)
            for event in step['wait']['digital']:
                self.client.write_holding(
                    4013+event['number']-1,
                    eventlookup.index(event['value']) if event['enable'] else 0
                )
            for analog in step['wait']['analog']:
                if analog['enable']:
                    reg = 4021+(analog['number']-1)*2
                    value = int(analog['value']/self.__get_scalar(analog['number']))
                    self.client.write_holding(reg, 1)
                    self.client.write_holding_signed(reg+1, value)
                else:
                    self.client.write_holding(4021+(analog['number']-1)*2, 0)
        else:
            self.client.write_holding(4012, 0)
        for event in step['events']:
            self.client.write_holding(4030+event['number']-1, 1 if event['value'] == 'on' else 0)
        if step['type'] != 'ramprate':
            dur = step['duration']
            self.client.write_holding(4009, [dur['hours'], dur['minutes'], dur['seconds']])
        else:
            self.client.write_holding(4043, int(step['loops'][0]['rate']*10))
        for i, loop in enumerate(step['loops']):
            if i < self.loops + self.cascades:
                target = int(loop['target']/self.__get_scalar(i+1))
                self.client.write_holding_signed(4044+i, target)#setpoint
                self.client.write_holding(4046+i, loop['pidset'] - 5*i - 1)#pid set
                self.client.write_holding(4048+i, 1 if loop['gsoak'] else 0) #guarrneteed soak
                if self.combined_event[i]:
                    reg = 4030+self.combined_event[i]-1
                    self.client.write_holding(reg, 1 if loop['enable'] else 0)
        if self.cond_event: #if we have a condition event force it on for the profile
            self.client.write_holding(4030+self.cond_event-1, 1)

    def __edit_prgm_step_jump(self, step):
        '''setup an jump step'''
        if step['jprofile'] == 0:
            profile_num = self.client.read_holding(4000)[0]
        else:
            profile_num = step['jprofile']
        self.client.write_holding(4003, 4)
        self.client.write_holding(4050, [profile_num, step['jstep'], step['jcount']])

    def __edit_prgm_step_end(self, step):
        '''setup an end step'''
        if step['action'] == 'hold':
            self.client.write_holding(4060, 0)
        elif step['action'] == 'controloff':
            self.client.write_holding(4060, 1)
        elif step['action'] == 'alloff':
            self.client.write_holding(4060, 2)
        elif step['action'] == 'idle':
            cache = [3]
            for idxl, loop in enumerate(step['loops']):
                cache.append(int(loop['target'] / self.__get_scalar(idxl+1)))
            self.client.write_holding_signed(4060, cache)
        else:
            raise ValueError('invalid end step action')

    def __edit_prgm_step(self, step):
        '''edit the selected step parameters (excluding an end step)'''
        if step['type'] == 'autostart':
            self.__edit_prgm_step_autostart(step)
        elif step['type'] in ['ramptime', 'ramprate', 'soak']:
            self.__edit_prgm_step_rampsoak(step)
        elif step['type'] == 'jump':
            self.__edit_prgm_step_jump(step)
        elif step['type'] == 'end':
            self.__edit_prgm_step_end(step)
        else:
            raise ValueError('invalid step type')

    def __create_prgm(self, value):
        '''create a new program return its number'''
        self.client.write_holding(4002, 1)
        num = self.client.read_holding(4000)[0]
        self.__set_prgm_name(num, value['name'])
        for i, step in enumerate(value['steps']):
            if step['type'] != 'end':
                self.client.write_holding(4001, [i+1, 2])
            else:
                self.client.write_holding(4001, i+1)
            self.__edit_prgm_step(step)
        return num

    def __edit_prgm(self, num, value):
        '''Edit an existing program'''
        self.prgm_delete(num, exclusive=False)
        return self.__create_prgm(value)


    def connect(self):
        '''
        connect to the controller using the paramters provided on class initialization
        '''
        if self.interface == "RTU":
            self.client = ModbusRTU(
                address=self.adr,
                port=self.serialport,
                baud=self.baudrate,
                timeout=10.0
            )
        else:
            self.client = ModbusTCP(self.adr, self.host, timeout=10.0)

    def close(self):
        '''
        close the connection to the controller
        '''
        try:
            self.client.close()
        except AttributeError:
            pass
        self.client = None

    @exclusive
    def raw(self, command):
        '''
        connect directly to the controller
        '''
        return self.client.interact(command)

    @exclusive
    def get_datetime(self):
        dtl = self.client.read_holding(1916, 6)
        return datetime.datetime(
            hour=dtl[0],
            minute=dtl[1],
            second=dtl[2],
            month=dtl[3],
            day=dtl[4],
            year=dtl[5]
        )

    @exclusive
    def set_datetime(self, value):
        vals = [value.hour, value.minute, value.second, value.month, value.day, value.year]
        self.client.write_holding(1916, vals)

    @exclusive
    def get_refrig(self):
        raise NotImplementedError

    @exclusive
    def set_refrig(self, value):
        raise NotImplementedError

    @exclusive
    def get_loop_sp(self, N, range_restrict=True):
        self.__range_check(N, 1, self.loops + self.cascades)
        minsp = self.client.read_holding_signed([602, 612][N-1])[0] * self.__get_scalar(N)
        const = self.client.read_holding_signed([300, 319][N-1])[0] * self.__get_scalar(N)
        if self.client.read_holding(200)[0] in [2, 3]: #running or holding profile
            cur = self.client.read_holding_signed([4122, 4123][N-1])[0] * self.__get_scalar(N)
        else:
            cur = const
        if range_restrict:
            return {
                'constant': const if const > minsp else minsp,
                'current': cur if cur > minsp else minsp
            }
        else:
            return {'constant': const, 'current': cur}

    @exclusive
    def set_loop_sp(self, N, value):
        self.__range_check(N, 1, self.loops + self.cascades)
        value = value['constant'] if isinstance(value, dict) else value
        value = int(value / self.__get_scalar(N)) #trim to 16bit signed int
        self.client.write_holding_signed([300, 319][N-1], value)

    @exclusive
    def get_loop_pv(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        return {'air': self.client.read_holding_signed([100, 104][N-1])[0] * self.__get_scalar(N)}

    @exclusive
    def get_loop_range(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        return {
            'max':self.client.read_holding_signed([603, 613, 623][N-1])[0] * self.__get_scalar(N),
            'min':self.client.read_holding_signed([602, 612, 622][N-1])[0] * self.__get_scalar(N)
        }

    @exclusive
    def set_loop_range(self, N, value):
        self.__range_check(N, 1, self.loops + self.cascades)
        vals = [int(value['min'] / self.__get_scalar(N)), int(value['max'] / self.__get_scalar(N))]
        self.client.write_holding_signed([602, 612, 622][N-1], vals)

    @exclusive
    def get_loop_en(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        lrange = self.get_loop_range(N, exclusive=False)
        profile = self.client.read_holding(200)[0] in [2, 3]
        cmd = self.get_loop_sp(N, False, exclusive=False)['current'] >= lrange['min']
        if self.combined_event[N-1] > 0:
            eve = self.get_event(self.combined_event[N-1], exclusive=False)['constant']
            running = self.get_event(self.cond_event, exclusive=False)['current'] if self.cond_event else True
            return {'constant': eve, 'current': eve if running else cmd}
        else:
            return {'constant': cmd, 'current': True if profile else cmd}

    @exclusive
    def set_loop_en(self, N, value):
        self.__range_check(N, 1, self.loops + self.cascades)
        value = value['constant'] if isinstance(value, dict) else value
        lrange = self.get_loop_range(N, exclusive=False)
        if self.get_loop_sp(N, exclusive=False)['constant'] < lrange['min'] and value:
            self.set_loop_sp(N, lrange['min'], exclusive=False)
        if self.combined_event[N-1] != 0:
            self.set_event(self.combined_event[N-1], value, exclusive=False)

    @exclusive
    def get_loop_units(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        units = ['Temp', u'%RH', u'PSI', u''][self.client.read_holding([608, 618, 628][N-1])[0]]
        if units == 'Temp':
            units = u'\xb0C' if self.client.read_holding(901)[0] else u'\xb0F'
        return units

    @exclusive
    def get_loop_mode(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        lpen = self.get_loop_en(N, exclusive=False)
        lpen['constant'] = 'On' if lpen['constant'] else 'Off'
        lpen['current'] = 'On' if lpen['current'] else 'Off'
        return lpen

    @exclusive
    def get_loop_modes(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        return ['On', 'Off'] if self.combined_event[N-1] != 0 else ['On']

    @exclusive
    def set_loop_mode(self, N, value):
        self.__range_check(N, 1, self.loops + self.cascades)
        value = value['constant'] if isinstance(value, dict) else value
        if value in ['On', 'Off']:
            self.set_loop_en(N, value == 'On', exclusive=False)
        else:
            raise ValueError('Unsupported Loop Mode, must be "On" or "Off" not "%r"' % value)

    @exclusive
    def get_loop_power(self, N):
        self.__range_check(N, 1, self.loops + self.cascades)
        outa = self.client.read_holding_signed([103, 111][N-1])[0] / 100.0
        outb = self.client.read_holding_signed([107, 115][N-1])[0] / 100.0
        combined = outa + outb
        return {'constant':combined, 'current':combined}

    @exclusive
    def set_loop_power(self, N, value):
        raise NotImplementedError

    @exclusive
    def get_cascade_sp(self, N):
        self.__range_check(N, 1, self.cascades)
        vals = self.get_loop_sp(1, exclusive=False)
        vals.update({
            'air':self.client.read_holding_signed(1922)[0]*self.__get_scalar(1),
            'product':vals['current']
        })
        return vals

    @exclusive
    def set_cascade_sp(self, N, value):
        self.__range_check(N, 1, self.cascades)
        return self.set_loop_sp(N, value, exclusive=False)

    @exclusive
    def get_cascade_pv(self, N):
        self.__range_check(N, 1, self.cascades)
        return {
            'air':self.client.read_holding_signed(100)[0] * self.__get_scalar(1),
            'product':self.client.read_holding_signed(108)[0] * self.__get_scalar(3)
        }

    @exclusive
    def get_cascade_range(self, N):
        self.__range_check(N, 1, self.cascades)
        return self.get_loop_range(N, exclusive=False)

    @exclusive
    def set_cascade_range(self, N, value):
        self.__range_check(N, 1, self.cascades)
        return self.set_loop_range(N, value, exclusive=False)

    @exclusive
    def get_cascade_en(self, N):
        self.__range_check(N, 1, self.cascades)
        return self.get_loop_en(N, exclusive=False)

    @exclusive
    def set_cascade_en(self, N, value):
        self.__range_check(N, 1, self.cascades)
        return self.set_loop_en(N, value, exclusive=False)

    @exclusive
    def get_cascade_units(self, N):
        self.__range_check(N, 1, self.cascades)
        return self.get_loop_units(N, exclusive=False)

    @exclusive
    def get_cascade_mode(self, N):
        self.__range_check(N, 1, self.cascades)
        return self.get_loop_mode(N, exclusive=False)

    @exclusive
    def get_cascade_modes(self, N):
        self.__range_check(N, 1, self.cascades)
        return self.get_loop_modes(N, exclusive=False)

    @exclusive
    def set_cascade_mode(self, N, value):
        self.__range_check(N, 1, self.cascades)
        return self.set_loop_mode(N, value, exclusive=False)

    @exclusive
    def get_cascade_ctl(self, N):
        self.__range_check(N, 1, self.cascades)
        return {'constant':True, 'current':True}

    @exclusive
    def set_cascade_ctl(self, N, value):
        raise NotImplementedError

    @exclusive
    def get_cascade_deviation(self, N):
        self.__range_check(N, 1, self.cascades)
        vals = self.client.read_holding_signed(1926, 2)
        return {
            'positive': vals[1] * self.__get_scalar(1),
            'negative': vals[0] * self.__get_scalar(1)
        }

    @exclusive
    def set_cascade_deviation(self, N, value):
        self.__range_check(N, 1, self.cascades)
        vals = [
            int(value['negative']/self.__get_scalar(1)),
            int(value['positive']/self.__get_scalar(1))
        ]
        self.client.write_holding_signed(1926, vals)
        self.client.write_holding(25, 0)

    @exclusive
    def get_cascade_power(self, N):
        self.__range_check(N, 1, self.cascades)
        return self.get_loop_power(N, exclusive=False)

    @exclusive
    def set_cascade_power(self, N, value):
        raise NotImplementedError

    @exclusive
    def get_event(self, N):
        self.__range_check(N, 1, 8)
        val = self.client.read_holding(2000 + 10*(N-1))[0] == 1
        return {'current':val, 'constant':val}

    @exclusive
    def set_event(self, N, value):
        self.__range_check(N, 1, 8)
        value = value['constant'] if isinstance(value, dict) else value
        self.client.write_holding(2000 + 10*(N-1), 1 if value else 0)

    @exclusive
    def get_status(self):
        if len(self.get_alarm_status(exclusive=False)['active']) > 0:
            return 'Alarm'
        mode = self.client.read_holding(200)[0]
        prof = ['Constant', 'Constant', 'Program Running', 'Program Paused'][mode]
        if self.cond_event:
            return prof if self.get_event(self.cond_event, exclusive=False)['current'] else 'Standby'
        else:
            return prof

    @exclusive
    def get_alarm_status(self):
        aalms, ialms = [], []
        for alm in self.limits:
            if self.__get_digital_input(abs(alm)): #this needs to be configurable
                ialms.append(alm)
            else:
                aalms.append(alm)
        return {'active': aalms, 'inactive': ialms}

    @exclusive
    def const_start(self):
        self.client.write_holding(1217, 1) #terminate profile
        if self.cond_event:
            self.set_event(self.cond_event, True, exclusive=False)

    @exclusive
    def stop(self):
        self.client.write_holding(1217, 1) #terminate profile
        if self.cond_event:
            self.set_event(self.cond_event, False, exclusive=False)

    @exclusive
    def prgm_start(self, N, step):
        self.client.write_holding(4000, N)
        self.client.write_holding(4001, step)
        self.client.write_holding(4002, 5)

    @exclusive
    def prgm_pause(self):
        self.client.write_holding(1210, 1)

    @exclusive
    def prgm_resume(self):
        self.client.write_holding(1209, 1)

    @exclusive
    def get_prgm_counter(self):
        try:
            status = self.client.read_holding(4126, 3) #count, profile, step
            self.client.write_holding(4000, (status[1:]))
            prof = self.client.read_holding(4050, 3)
            return [
                {
                    'name':str(status[0]),
                    'start':prof[1],
                    'end':status[2],
                    'cycles':prof[2],
                    'remaining':prof[2]-status[0]
                }
            ]
        except Exception:
            return []

    @exclusive
    def prgm_next_step(self):
        program = self.get_prgm_cur(exclusive=False)
        nextstep = self.get_prgm_cstep(exclusive=False) + 1
        self.const_start(exclusive=False)
        time.sleep(2)
        self.prgm_start(program, nextstep, exclusive=False)

    @exclusive
    def get_prgm_cur(self):
        return self.client.read_holding(4100)[0]

    @exclusive
    def get_prgm_cstep(self):
        return self.client.read_holding(4101)[0]

    @exclusive
    def get_prgm_cstime(self):
        data = self.client.read_holding(4119, 3)
        return "%d:%02d:%02d" % (data[0], data[1], data[2])

    @exclusive
    def get_prgm_time(self, pgm=None):
        rpgmi = self.get_prgm_cur(exclusive=False)
        if pgm is None:
            pgm = self.get_prgm(rpgmi, exclusive=False)
        jumps = 0
        for step in pgm['steps']:
            jumps += 1 if step['type'] == 'jump' else 0
            if jumps > 1:
                return 'ERROR:jump qty'
            if step['type'] == 'jump' and step.get('jprofile', rpgmi) != 0:
                return 'ERROR:jump prfl'
        rtime = list(self.client.read_holding(4119, 3)) #[hours,minutes,seconds]
        jump_cnt = self.client.read_holding(4126)[0]
        stepi = self.get_prgm_cstep(exclusive=False) + 1
        ptarget = None
        subtractor = 2
        while ptarget is None and stepi - subtractor > 0:
            if pgm['steps'][stepi-subtractor]['type'] in ['ramprate', 'ramptime', 'soak']:
                ptarget = pgm['steps'][stepi-subtractor]['loops'][0]['target']
            subtractor -= 1
        if ptarget is None:
            ptarget = 0 #useless default
        while stepi < len(pgm['steps']):
            step = pgm['steps'][stepi-1]
            if step['type'] in ['ramptime', 'soak']:
                rtime[0] += step['duration']['hours']
                rtime[1] += step['duration']['minutes']
                rtime[2] += step['duration']['seconds']
                ptarget = step['loops'][0]['target']
            elif step['type'] == 'ramprate':
                diff = abs(step['loops'][0]['target'] - ptarget)
                minutes = diff/step['loops'][0]['rate']
                rtime[1] += int(minutes)
                rtime[2] += int(minutes%60)
                ptarget = step['loops'][0]['target']
            if step['type'] == 'jump' and jump_cnt < step['jcount']:
                jump_cnt += 1
                stepi = step['jstep']
            else:
                stepi += 1
        rtime[1] += int(rtime[2]/60)
        rtime[2] = int(rtime[2]%60)
        rtime[0] += int(rtime[2]/60)
        rtime[1] = int(rtime[1]%60)
        return '%d:%02d:%02d' % tuple(rtime)

    @exclusive
    def get_prgm_name(self, N):
        parsed = re.search(r'([A-Z0-9 ]+)', self.client.read_holding_string(3500+10*(N-1), 10))
        return parsed.group(1).rstrip()

    @exclusive
    def set_prgm_name(self, N, value):
        self.__set_prgm_name(N, value)
        self.client.write_holding(25, 0) #save changes to profiles

    @exclusive
    def get_prgm_steps(self, N):
        self.__range_check(N, 1, 41)
        self.client.write_holding(4000, N)
        for step in range(2, 257):
            self.client.write_holding(4001, step)
            step_type = self.client.read_holding(4003)[0]
            if step_type == 5:
                return step
            elif step_type > 5:
                raise ValueError('Program #%d does not exist' % N)

    @exclusive
    def get_prgms(self):
        programs = []
        num_programs = 40 - self.client.read_holding(1218)[0]
        for idx in range(1, 41):
            self.client.write_holding(4001, 2)
            if num_programs > 0:
                self.client.write_holding(4000, [idx, 1])
                if self.client.read_holding(4003)[0] <= 5:
                    programs.append({'number':idx, 'name':self.get_prgm_name(idx, exclusive=False)})
                    num_programs -= 1
                else:
                    programs.append({'number': idx, 'name': ''})
            else:
                programs.append({'number': idx, 'name': ''})
        return programs

    def __get_prgm_empty(self):
        '''
        create an empty null program
        '''
        dins = [self.client.read_holding(1060+i*2)[0] == 10 for i in range(4)]
        analogs = self.__get_analog_input_setup()
        ret = {
            'name':'PROFILE',
            'steps_available': self.client.read_holding(1219)[0],
            'steps':[{
                'type':'end',
                'start_type':'day',
                'start_date':{'month':1, 'day':1, 'year':2000},
                'start_time':{'day':'sunday', 'hours':0, 'minutes':0, 'seconds':0},
                'loops':[
                    {
                        'enable':True,
                        'target':0,
                        'pidset':1 + 5*i,
                        'gsoak':False,
                        'isCascade': self.cascades > 0 if i == 0 else False,
                        'showEnable': self.combined_event[i] > 0
                    }
                    for i in range(self.loops + self.cascades)
                ],
                'events':[
                    {'number':i, 'value':'off'}
                    for i in range(1, 9) if i not in self.combined_event and i != self.cond_event
                ],
                'wait':{
                    'enable':False,
                    'digital':[
                        {'number':i+1, 'enable':False, 'value':'off'}
                        for i in range(4) if dins[i]
                    ],
                    'analog':[
                        {'number':i+1, 'enable':False, 'value':0.0}
                        for i in range(len(analogs)) if analogs[i] != 'off'
                    ]
                },
                'duration':{'hours':0, 'minutes':0, 'seconds':0},
                'jprofile':0,
                'jstep':0,
                'jcount':0,
                'action':'hold'
            }]
        }
        if self.loops + self.cascades == 1:
            ret['steps'][0]['loops'][0]['rate'] = 0.0
        return ret


    @exclusive
    def get_prgm(self, N):
        if N == 0:
            return self.__get_prgm_empty()
        rbase = 4003 #first register to read
        self.__range_check(N, 1, 40)
        self.client.write_holding(4000, N)
        program = {
            'steps':[],
            'name': self.get_prgm_name(N, exclusive=False),
            'steps_available': self.client.read_holding(1219)[0]
        }
        daylookup = ['daily', 'sun', 'mon', 'tue', 'wed', 'thur', 'fri', 'sat']
        dins = [self.client.read_holding(1060+i*2)[0] == 10 for i in range(4)]
        analogs = self.__get_analog_input_setup()
        for step in range(1, 257):
            self.client.write_holding(4001, step)
            params = self.client.read_holding_signed(rbase, 4062-rbase+1) #registers 4003-4062
            step_params = {
                'type':'soak',
                'start_type':'day',
                'start_date':{'month':1, 'day':1, 'year':2000},
                'start_time':{'day':'sunday', 'hours':0, 'minutes':0, 'seconds':0},
                'loops':[
                    {
                        'enable':True,
                        'target':0,
                        'pidset':1 + 5*i,
                        'gsoak':False,
                        'isCascade': self.cascades > 0 if i == 0 else False,
                        'showEnable': self.combined_event[i] > 0
                    }
                    for i in range(self.loops + self.cascades)
                ],
                'events':[
                    {'number':i, 'value':'off'}
                    for i in range(1, 9) if i not in self.combined_event and i != self.cond_event
                ],
                'wait':{
                    'enable':False,
                    'digital':[
                        {'number':i+1, 'enable':False, 'value':'off'}
                        for i in range(4) if dins[i]
                    ],
                    'analog':[
                        {'number':i+1, 'enable':False, 'value':0.0}
                        for i in range(len(analogs)) if analogs[i] != 'off'
                    ]
                },
                'duration':{'hours':0, 'minutes':0, 'seconds':0},
                'jprofile':0,
                'jstep':0,
                'jcount':0,
                'action':'hold'
            }
            if self.loops + self.cascades == 1:
                step_params['loops'][0]['rate'] = 0
            if params[4003-rbase] == 0:#Auto start
                step_params.update({
                    'type':'autostart',
                    'start_type':['date', 'day'][params[4004-rbase]],
                    'start_date':{
                        'month':params[4005-rbase],
                        'day':params[4006-rbase],
                        'year':params[4007-rbase]
                    },
                    'start_time':{
                        'day':daylookup[params[4008-rbase]],
                        'hours':params[4009-rbase],
                        'minutes':params[4010-rbase],
                        'seconds':params[4011-rbase]
                    }
                })
            elif params[4003-rbase] in [1, 2, 3]: #ramptime, ramprate, soak
                step_params.update({
                    'type': ['', 'ramptime', 'ramprate', 'soak'][params[4003-rbase]],
                    'wait':{'enable': params[4012-rbase] == 1},
                    'events':[
                        {'number':i, 'value':'on' if params[4030-rbase-1+i] else 'off'}
                        for i in range(1, 9) if i not in self.combined_event and i != self.cond_event
                    ],
                    'loops':[
                        {
                            'enable':params[4030-rbase-1+self.combined_event[i]] == 1 \
                                        if self.combined_event[i] else True,
                            'target':params[4044-rbase+i] * self.__get_scalar(i+1),
                            'pidset':params[4046-rbase+i] + 1 + 5*i,
                            'gsoak':params[4048-rbase+i] == 1,
                            'isCascade': self.cascades > 0 if i == 0 else False,
                            'showEnable': self.combined_event[i] > 0
                        }
                        for i in range(self.loops + self.cascades)
                    ]
                })
                if self.loops + self.cascades == 1:
                    step_params['loops'][0]['rate'] = 0
                if params[4003-rbase] != 2:
                    step_params['duration'] = {
                        'hours':params[4009-rbase],
                        'minutes':params[4010-rbase],
                        'seconds':params[4011-rbase]
                    }
                else:
                    step_params['loops'][0]['rate'] = params[4043-rbase] / 10.0
                step_params['wait'].update({
                    'digital':[
                        {
                            'number':i+1,
                            'enable':params[4013-rbase+i] != 0,
                            'value':['off', 'off', 'on'][params[4013-rbase+i]]
                        }
                        for i in range(4) if dins[i]
                    ],
                    'analog':[
                        {
                            'number':i+1,
                            'enable':params[4021-rbase+i*2] == 1,
                            'value':params[4022-rbase+i*2] * self.__get_scalar(i+1)
                        }
                        for i in range(len(analogs)) if analogs[i] != 'off'
                    ]
                })
            elif params[4003-rbase] == 4:#Jump
                #if we are jumping to the same profile hprofile will be 0
                prof_num = params[4050-rbase]
                prof_num = 0 if prof_num == N else prof_num
                step_params.update({
                    'type':'jump',
                    'jprofile':prof_num,
                    'jstep':params[4051-rbase],
                    'jcount':params[4052-rbase]
                })
            elif params[4003-rbase] == 5:#End
                step_params.update({
                    'type':'end',
                    'action':['hold', 'controloff', 'alloff', 'idle'][params[4060-rbase]],
                    'loops':[
                        {
                            'enable':True,
                            'target':params[4061-rbase+i]*self.__get_scalar(i+1),
                            'pidset':1 + 5*i,
                            'gsoak':False,
                            'isCascade': self.cascades > 0 if i == 0 else False,
                            'showEnable': self.combined_event[i] > 0
                        }
                        for i in range(self.loops + self.cascades)
                    ]
                })
                if self.loops + self.cascades == 1:
                    step_params['loops'][0]['rate'] = 0
            else:#Not a program
                raise ControllerInterfaceError('Program #%d does not exist' % N)
            program['steps'].append(step_params)
            if step_params['type'] == 'end':
                break
        return program

    @exclusive
    def set_prgm(self, N, value):
        self.__range_check(N, 1, 40)
        try:
            self.get_prgm_steps(N, exclusive=False)
            self.__edit_prgm(N, value)
        except ValueError:
            N = self.__create_prgm(value)
        self.client.write_holding(25, 0) #save the new program
        return N

    @exclusive
    def prgm_delete(self, N):
        self.__range_check(N, 1, 40)
        self.client.write_holding(4000, [N, 1, 3]) #program #, step 1, action delete
        self.client.write_holding(25, 0) #save changes to profiles


    @exclusive
    def process_controller(self, update=True):
        part_no = 'Watlow F4'
        if update:
            self.loops = 1
            self.cascades = 0
        try:
            self.client.write_holding(319, self.client.read_holding(319)[0])
            if update:
                self.loops = 2
            part_no += 'D'
        except Exception:
            part_no += 'S'
        try:
            if self.client.read_holding(1925)[0]:
                part_no += ' W/Cascade'
                if update:
                    self.loops -= 1
                    self.cascades = 1
        except Exception:
            pass
        self.__update_loop_map()
        self.combined_event = [
            self.cascade_event[0] if self.cascades > 0 else self.loop_event[0],
        ]
        if self.loops == 2:
            self.combined_event.append(self.loop_event[1])
        elif self.cascades == 1 and self.loops == 1:
            self.combined_event.append(self.loop_event[0])
        return part_no

    @exclusive
    def get_network_settings(self):
        raise NotImplementedError

    @exclusive
    def set_network_settings(self, value):
        raise NotImplementedError

    def get_operation_modes(self):
        if self.cond_event is None:
            return ['constant', 'program']
        else:
            return ['standby', 'constant', 'program']



#!/usr/bin/python
'''
A simple command line program that calls all functions supported by the
chamberconnectlibrary.

:copyright: (C) Espec North America, INC.
:license: MIT, see LICENSE for more details.
'''
#pylint: disable=W0703
import sys
import traceback
from chamberconnectlibrary.espec import Espec
from chamberconnectlibrary.watlowf4t import WatlowF4T
from chamberconnectlibrary.watlowf4 import WatlowF4

def main(**kwargs):
    '''Try each command given a set of parameters'''
    ctlr_type = kwargs.pop('controller', None)
    if ctlr_type == 'Espec' or ctlr_type == 'EspecP300':
        ctlr = Espec(**kwargs)
    elif ctlr_type == 'EspecSCP220':
        ctlr = Espec(ctlr_type='SCP220', **kwargs)
    elif ctlr_type == 'WatlowF4':
        ctlr = WatlowF4(**kwargs)
    else:
        ctlr = WatlowF4T(**kwargs)
    ctlr.process_controller()
    ctlr.self_test(ctlr.loops+ctlr.cascades, ctlr.cascades)

if __name__ == '__main__':
    try:
        main(
            controller=sys.argv[1],
            interface=sys.argv[2],
            serialport=sys.argv[3],
            host=sys.argv[3],
            baudrate=int(sys.argv[4]) if len(sys.argv) == 5 else 9600
        )
    except Exception:
        traceback.print_exc()

        print '\nThe test could not be run try:'
        print '\nchamberconnectlibrary_test controller interface ipORserialport [baudrate]'
        print '\tcontroller: "Espec"/"EspecP300", "EspecSCP220", "WatlowF4", or "WatlowF4T"(default)'
        print '\tinterface: "Serial": Serial connection when "controller" is "Espec".'
        print '\t           "RTU":Serial connection when "controller" is "WatlowF4T"'
        print '\t           "TCP":TCP connection'
        print '\t"hostORserialport": hostname for TCP, or serial port for RTU/Serial'
        print '\t"baudrate": The baudrate for RTU/Serial, optional(default=9600)'



'''
Common interface for all All ChamberConnectLibrary upper level interfaces

:copyright: (C) Espec North America, INC.
:license: MIT, see LICENSE for more details.
'''
#pylint: disable=W0703,W0201,R0902,W0232,R0904,C0103
from abc import ABCMeta, abstractmethod
from threading import RLock
import traceback
import time
import inspect

class ControllerInterfaceError(Exception):
    '''Exception that is thrown when a there is a problem communicating with a controller'''
    pass

def exclusive(func):
    '''Lock the physical interface for the function call'''
    def wrapper(self, *args, **kwargs):
        '''Lock the physical interface for the function call'''
        if kwargs.pop('exclusive', True):
            with self.lock:
                auto_connect = self.client is None
                try:
                    if auto_connect:
                        self.connect()
                    return func(self, *args, **kwargs)
                finally:
                    try:
                        if auto_connect:
                            self.close()
                    except Exception:
                        pass
        else:
            try:
                del kwargs['exclusive']
            except Exception:
                pass
            return func(self, *args, **kwargs)
    return wrapper

class ControllerInterface:
    '''Abstract class for a controller implimentation of the chamberconnectlibrary'''
    __metaclass__ = ABCMeta

    loop_map = []
    named_loop_map = {}

    def init_common(self, **kwargs):
        '''Setup properties of all controllers of the chamberconnectlibrary'''
        self.client = None
        self.host = kwargs.get('host')
        self.interface = kwargs.get('interface')
        self.adr = kwargs.get('adr', 1)
        self.serialport = kwargs.get('serialport')
        self.baudrate = kwargs.get('baudrate')
        self.loops = kwargs.get('loops', 1)
        self.cascades = kwargs.get('cascades', 0)
        self.lock = kwargs.get('lock', RLock())

    @abstractmethod
    def get_datetime(self):
        '''
        Get the datetime from the controller.

        Returns:
            datetime. The current time as a datetime.datetime object
        '''
        pass

    @abstractmethod
    def set_datetime(self, value):
        '''
        Set the datetime of the controller.

        Args:
            value (datetime.datetime): The new datetime for the controller.
        '''
        pass

    @abstractmethod
    def get_refrig(self):
        '''
        Get the constant settings for the refigeration system

        returns:
            {"mode":string, "setpoint":float}
        '''
        pass

    @abstractmethod
    def set_refrig(self, value):
        '''
        Set the constant setpoints refrig mode

        params:
            mode: string,"off" or "manual" or "auto"
            setpoint: int,20 or 50 or 100
        '''
        pass

    @exclusive
    def get_loop(self, identifier, *args):
        '''
        Get all parameters for a loop from a given list.
        There are four different ways to call this method; all methods return the same data:

        get_loop(str(identifier), *str(parameters))
            Args:
                identifier (str): The name of the loop.
                parameters (list(string)): The list of parameters to get from the loop, (see below)

        get_loop(str(identifier), [str(parameters)])
            Args:
                identifier (str): The name of the loop.
                parameters (list(string)): The list of parameters to get from the loop, (see below)

        get_loop(int(identifier), str(loop_type), *str(parameters))
            Args:
                identifier (str): The name of the loop.
                loop_type (str): The type of loop to be accessed ("loop" or "cascade")
                parameters (list(string)): The list of parameters to get from the loop, (see below)

        get_loop(int(identifier), str(loop_type), [str(parameters)])
            Args:
                identifier (str): The name of the loop.
                loop_type (str): The type of loop to be accessed ("loop" or "cascade")
                parameters (list(string)): The list of parameters to get from the loop, (see below)

        parameters:
            The following is a list of available parameters as referenced by each call type:
                "setpoint" -- The target temp/humi/altitude/etc of the control loop
                "processvalue" -- The current conditions inside the chamber
                "range" -- The settable range for the "setpoint"
                "enable" -- Weather the loop is on or off
                "units" -- The units of the "setpoint" or "processvalue" parameters
                "mode" -- The current control status of the loop
                "power" -- The current output power of the loop
                "deviation" -- (type="cascade" only) The allowable difference between air/prod.
                "enable_cascade" -- (type="cascade" only) Enable or disable cascade type control
        Returns:
            dict. The dictionary contains a key for each item in the list argument::
                "setpoint" -- {"constant":float,"current":float}
                "processvalue" -- {"air":float,"product":float} ("product" only w/ type="cascade")
                "range" -- {"min":float,"max":float}
                "enable" -- {"constant":bool,"current":bool}
                "units" -- str
                "mode" -- str('Off' or 'Auto' or 'Manual')
                "deviation" -- {"positive": float, "negative": float}
                "enable_cascade" -- {"constant":bool,"current":bool}
                "power" -- {"constant": float, "current": float}
        '''
        loop_functions = {
            'cascade':{
                'setpoint':self.get_cascade_sp,
                'setPoint':self.get_cascade_sp,
                'setValue':self.get_cascade_sp,
                'processvalue':self.get_cascade_pv,
                'processValue':self.get_cascade_pv,
                'range':self.get_cascade_range,
                'enable':self.get_cascade_en,
                'units':self.get_cascade_units,
                'mode':self.get_cascade_mode,
                'deviation':self.get_cascade_deviation,
                'enable_cascade':self.get_cascade_ctl,
                'power': self.get_cascade_power
            },
            'loop':{
                'setpoint':self.get_loop_sp,
                'setPoint':self.get_loop_sp,
                'setValue':self.get_loop_sp,
                'processvalue':self.get_loop_pv,
                'processValue':self.get_loop_pv,
                'range':self.get_loop_range,
                'enable':self.get_loop_en,
                'units':self.get_loop_units,
                'mode':self.get_loop_mode,
                'power':self.get_loop_power
            }
        }

        if isinstance(identifier, basestring):
            my_loop_map = self.loop_map[self.named_loop_map[identifier]]
            loop_number = my_loop_map['num']
            loop_type = my_loop_map['type']
            param_list = args if len(args) > 0 else None
        elif isinstance(identifier, int) and len(args) >= 1:
            loop_number = identifier
            loop_type = args[0]
            param_list = args[1:] if len(args) > 1 else None
        else:
            raise ValueError(
                'invalid argument format, call w/: '
                'get_loop(int(identifier), str(loop_type), *args) or '
                'get_loop(str(identifier), *args), *args are optional.'
            )

        if param_list is None:
            param_list = loop_functions[loop_type].keys()
            excludes = ['setPoint', 'setValue', 'processValue']
            param_list = [x for x in param_list if x not in excludes]
        elif len(param_list) >= 1 and \
             (isinstance(param_list[0], list) or isinstance(param_list[0], tuple)):
            param_list = param_list[0]
        ret = {}
        for key in param_list:
            try:
                ret[key] = loop_functions[loop_type][key](loop_number, exclusive=False)
            except KeyError:
                ret[key] = None
            except NotImplementedError:
                ret[key] = None
        return ret

    @exclusive
    def set_loop(self, identifier, loop_type='loop', param_list=None, **kwargs):
        '''
        Set all parameters for a loop from a given list.

        Args:
            identifier (int or str): The loop number, or the name of the loop
            loop_type (str): The loop type (disregarded when identifier is a str)::
                "cascade" -- A cascade control loop.
                "loop" -- A standard control loop (default).
            param_list (dict(dict)): The parameters to update as a dictionary::
                see kwargs for possible keys/values
            kwargs (dict): The parameters to update as key word arguments, param_list overrides::
                "setpoint" -- The target temp/humi/altitude/etc of the control loop
                "range" -- The settable range for the "setpoint"
                "enable" -- turn the control loop on or off
                "power" -- set the manual power of the control loop
                "mode" -- set the control mode of the control loop
                "deviation" -- (type="cascade" only) The allowable difference between air/prod.
                "enable_cascade" -- (type="cascade" only) Enable or disable cascade type control
        Returns:
            None
        '''
        loop_functions = {
            'cascade':{
                'setpoint':self.set_cascade_sp,
                'setPoint':self.set_cascade_sp,
                'setValue':self.set_cascade_sp,
                'range':self.set_cascade_range,
                'enable':self.set_cascade_en,
                'deviation':self.set_cascade_deviation,
                'enable_cascade':self.set_cascade_ctl,
                'mode':self.set_cascade_mode,
                'power':self.set_cascade_power
            },
            'loop':{
                'setpoint':self.set_loop_sp,
                'setPoint':self.set_loop_sp,
                'setValue':self.set_loop_sp,
                'range':self.set_loop_range,
                'enable':self.set_loop_en,
                'mode':self.set_loop_mode,
                'power':self.set_loop_power
            }
        }
        if param_list is None:
            param_list = kwargs
        if isinstance(identifier, basestring):
            my_loop_map = self.loop_map[self.named_loop_map[identifier]]
            loop_number = my_loop_map['num']
            loop_type = my_loop_map['type']
        elif isinstance(identifier, (int, long)):
            loop_number = identifier
        else:
            raise ValueError(
                'invalid argument format, call w/: '
                'set_loop(int(identifier), str(loop_type), **kwargs) or '
                'get_loop(str(identifier), **kwargs)'
            )

        #mode must be done first
        if 'mode' in param_list:
            loop_functions[loop_type]['mode'](
                exclusive=False,
                N=loop_number,
                value=param_list.pop('mode')
            )
        for key, val in param_list.items():
            try:
                loop_functions[loop_type][key](
                    exclusive=False,
                    N=loop_number,
                    value=val
                )
            except KeyError:
                pass
            except NotImplementedError:
                pass

    @exclusive
    def get_operation(self, pgm=None):
        '''
        Get the complete operation status of the chamber (excludes control loops and events).

        Args:
            pgm (dict): A cached version of the running program, if None it will be retrieved.
        Returns:
            dict::
                "mode" -- str('program' or 'constant' or 'standby' or 'off' or 'alarm' or 'paused')
                "status" -- str(varies by controller; more detailed/format'd version of "mode" key)
                "program" --
                "alarm" -- [int] list of active alarms by number, empty list if no alarms
        '''
        status = self.get_status(exclusive=False)
        ret = {'status': status}
        if 'Paused' in status:
            ret['mode'] = 'program_pause'
        elif status.startswith('Prog'):
            ret['mode'] = 'program'
        elif status.startswith('Const'):
            ret['mode'] = 'constant'
        elif status.startswith('Stand'):
            ret['mode'] = 'standby'
        elif status == 'Off':
            ret['mode'] = 'off'
        elif status == 'Alarm':
            ret['mode'] = 'alarm'
        else:
            ret['mode'] = 'ERROR'

        if 'Program' in status:
            pnum = self.get_prgm_cur(exclusive=False)
            ret['program'] = {
                'number':pnum,
                'step':self.get_prgm_cstep(),
                'time_remaining':self.get_prgm_time(pgm, exclusive=False),
                'step_time_remaining':self.get_prgm_cstime(exclusive=False),
                'name':self.get_prgm_name(pnum, exclusive=False),
                'steps':self.get_prgm_steps(pnum, exclusive=False),
                'cycles':self.get_prgm_counter(exclusive=False)
            }
        else:
            ret['program'] = None

        if status == 'Alarm':
            ret['alarms'] = self.get_alarm_status(exclusive=False)['active']
        else:
            ret['alarms'] = None

        return ret


    @exclusive
    def set_operation(self, mode, **kwargs):
        '''
        Update the controllers operation mode (ie run a program, stop, constant etc)

        Args:
            mode (string): The mode to run:
                'standby' or 'off': Stop the chamber from running
                'constant': Start the configured constant mode
                'program': run a program
                'program_pause': pause the running program
                'program_resume': resume the paused program
                'program_advance': force the program to the next step.
        kwargs:
            program (int or dict): the program # to run, valid dict: {'number':int, 'step': int}
            step (int): the step # to start the program on. (Default=1)
        Returns:
            dict::
                "mode" -- str('program' or 'constant' or 'standby' or 'off' or 'alarm' or 'paused')
                "status" -- str(varies by controller; more detailed/format'd version of "mode" key)
                "program" --
                "alarm" -- [int] list of active alarms by number, empty list if no alarms
        '''
        if mode in ['standby', 'off']:
            self.stop(exclusive=False)
        elif mode == 'constant':
            self.const_start(exclusive=False)
        elif mode == 'program':
            if isinstance(kwargs['program'], dict):
                self.prgm_start(
                    kwargs['program']['number'],
                    kwargs['program'].get('step', 1),
                    exclusive=False
                )
            else:
                self.prgm_start(kwargs['program'], kwargs.get('step', 1), exclusive=False)
        elif mode == 'program_pause':
            self.prgm_pause(exclusive=False)
        elif mode == 'program_resume':
            self.prgm_resume(exclusive=False)
        elif mode == 'program_advance':
            self.prgm_next_step(exclusive=False)
        else:
            raise ValueError('unsupported mode parameter')

    @exclusive
    def get_program(self, N):
        '''
        Get a program (alias for get_prgm)

        Args:
            N (int): The program number
        Returns:
            dict (format varies from controller to controller)
        '''
        return self.get_prgm(N, exclusive=False)

    @exclusive
    def set_program(self, N, value):
        '''
        Set a program (alias for set_prgm)

        Args:
            N (int): The program number
            value (dict): The program to write to the controller, None erases the given program
        '''
        if value is None:
            return self.prgm_delete(N, exclusive=False)
        else:
            return self.set_prgm(N, value, exclusive=False)

    @exclusive
    def get_program_list(self):
        '''
        Get a list of all programs on the chamber. (alias for get_prgms)

        Returns:
            [{"number":int, "name":str}]
        '''
        return self.get_prgms(exclusive=False)

    @exclusive
    def get_program_details(self, N):
        '''
        Get the name and number of steps of a program.

        Args:
            N (int): The program number
        Returns:
            {"number":int, "name":str, "steps":int}
        '''
        return {
            'number':N,
            'name':self.get_prgm_name(N, exclusive=False),
            'steps':self.get_prgm_steps(N, exclusive=False)
        }

    @abstractmethod
    def get_loop_sp(self, N):
        '''
        Get the setpoint of a control loop.

        Args:
            N (int): The number for the control loop
            constant (bool): Read the constant or current setpoint, None=Both (default=None)
        Returns:
            {"constant":float, "current":float}
        Raises:
            ValueError
        '''
        pass

    @abstractmethod
    def set_loop_sp(self, N, value):
        '''
        Set the setpoint of the control loop.

        Args:
            N (int): The number for the control loop
            value (float): The new setpoint
        Returns:
            None
        Raises:
            ValueError
        '''
        pass

    @abstractmethod
    def get_loop_pv(self, N):
        '''
        Get the process value of a loop.

        Args:
            N (int): The number for the control loop
            product (bool): True=(not valid), False=air temp, None=both (default=None)
        Returns:
            dict(float). product=None
            float. product=bool
        Raises:
            ValueError
        '''
        pass

    @abstractmethod
    def get_loop_range(self, N):
        '''
        Get the valid setpoint range of a loop

        Args:
            N (int): The number of the loop
        Returns:
            {"min": float, "max": float}
        '''
        pass

    @abstractmethod
    def set_loop_range(self, N, value):
        '''
        Set the valid setpoint range of a loop

        Args:
            N (int): The number of the loop
            value: ({"min": float, "max": float}): The range
        '''
        pass

    @abstractmethod
    def get_loop_en(self, N):
        '''
        Get the enable/disable state of a loop

        Args:
            N (int): The number of the loop
        Returns:
            {"constant": bool, "current": bool}
        '''
        pass

    @abstractmethod
    def set_loop_en(self, N, value):
        '''
        Set the enable/disable state of a loop

        Args:
            N (int): The number of the loop
            value (bool): True = loop is running
        '''
        pass

    @abstractmethod
    def get_loop_units(self, N):
        '''
        Get the units for a loop

        Args:
            N (int): The number of the loop
        Returns:
            string
        '''
        pass

    @abstractmethod
    def get_loop_mode(self, N):
        '''
        Get the control mode for a loop

        Args:
            N (int): The number of the loop
        Returns:
            string: The control mode (varies by controller)
        '''
        pass

    @abstractmethod
    def get_loop_modes(self, N):
        '''
        Get the available modes for a loop

        Args:
            N (int): The number of the loop
        Returns:
            string: list of control modes
        '''
        pass

    @abstractmethod
    def set_loop_mode(self, N, value):
        '''
        Get the control mode for a loop

        Args:
            N (int): The number of the loop
            value (bool): The control mode (varies by controller)
        '''
        pass

    @abstractmethod
    def get_loop_power(self, N):
        '''
        Get the output power(%) for a loop

        Args:
            N (int): The number of the loop
        Returns:
            {"constant": float, "current": float}
        '''
        pass

    @abstractmethod
    def set_loop_power(self, N, value):
        '''
        Set the output power(%) for a loop

        Args:
            N (int): The number of the loop
            value (float): The output power
        '''
        pass

    @abstractmethod
    def get_cascade_sp(self, N):
        '''
        Get the setpoint for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"constant":float, "current":float, "air":float, "product":float}
        '''
        pass

    @abstractmethod
    def set_cascade_sp(self, N, value):
        '''
        Get the setpoint for a cascade loop

        Args:
            N (int): The number of the loop
            value (float): The setpoint
        '''
        pass

    @abstractmethod
    def get_cascade_pv(self, N):
        '''
        Get the process value for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"product":float, "air":float}
        '''
        pass

    @abstractmethod
    def get_cascade_range(self, N):
        '''
        Get the valid setpoint range for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"min":float, "max":float}
        '''
        pass

    @abstractmethod
    def set_cascade_range(self, N, value):
        '''
        Set the valid setpoint range for a cascade loop

        Args:
            N (int): The number of the loop
            value ({"min":float, "max":float}): The range
        '''
        pass

    @abstractmethod
    def get_cascade_en(self, N):
        '''
        Get the enable/disable state for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"constant":bool, "current":bool}
        '''
        pass

    @abstractmethod
    def set_cascade_en(self, N, value):
        '''
        Set the enable/disable state for a cascade loop

        Args:
            N (int): The number of the loop
            value (bool): True = loop running
        '''
        pass

    @abstractmethod
    def get_cascade_units(self, N):
        '''
        Get the units for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            str: The loop units
        '''
        pass

    @abstractmethod
    def get_cascade_mode(self, N):
        '''
        Get the control mode for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            str: The control mode
        '''
        pass

    @abstractmethod
    def get_cascade_modes(self, N):
        '''
        Get the available modes for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            string: list of control modes
        '''
        pass

    @abstractmethod
    def set_cascade_mode(self, N, value):
        '''
        Set the control mode for a cascade loop

        Args:
            N (int): The number of the loop
            value (str): The control mode
        '''
        pass

    @abstractmethod
    def get_cascade_ctl(self, N):
        '''
        Get enable/disable of cascade mode for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"constant":bool, "current":bool}
        '''
        pass

    @abstractmethod
    def set_cascade_ctl(self, N, value):
        '''
        Set enable/disable of cascade mode for a cascade loop

        Args:
            N (int): The number of the loop
            value (bool): True = when cascade mode is enabled
        '''
        pass

    @abstractmethod
    def get_cascade_deviation(self, N):
        '''
        Get allowable product to air deviation for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"positive":float, "negative":float}
        '''
        pass

    @abstractmethod
    def set_cascade_deviation(self, N, value):
        '''
        Set allowable product to air deviation for a cascade loop

        Args:
            N (int): The number of the loop
            value ({"positive": float, "negative": float}): The deviations
        '''
        pass

    @abstractmethod
    def get_cascade_power(self, N):
        '''
        Get the output power for a cascade loop

        Args:
            N (int): The number of the loop
        Returns:
            {"constant":float, "current":float}
        '''
        pass

    @abstractmethod
    def set_cascade_power(self, N, value):
        '''
        Set the output power for a cascade loop

        Args:
            N (int): The number of the loop
            value (float): The output power %
        '''
        pass

    @abstractmethod
    def get_event(self, N):
        '''
        Get the state of a programmable output

        Args:
            N (int): The number of the output
        Returns:
            {"constant":bool, "current":bool}
        '''
        pass

    @abstractmethod
    def set_event(self, N, value):
        '''
        Set the state of a programmable output

        Args:
            N (int): The number of the output
            value (bool): the output state True = on
        '''
        pass

    @abstractmethod
    def get_status(self):
        '''
        Get the chamber status.

        Returns:
            str: The chamber status
        '''
        pass

    @abstractmethod
    def get_alarm_status(self):
        '''
        Get the chamber alarms status.

        Returns:
            {"active":[int], "inactive":[int]}
        '''
        pass

    @abstractmethod
    def const_start(self):
        '''
        Start the constant mode of the chamber
        '''
        pass

    @abstractmethod
    def stop(self):
        '''
        Start operation of the chamber.
        '''
        pass

    @abstractmethod
    def prgm_start(self, N, step):
        '''
        Start a program on the chamber

        Args:
            N (int): The number of the program to start.
            step (int): The step to start the program on.
        '''
        pass

    @abstractmethod
    def prgm_pause(self):
        '''
        Pause the running program.
        '''
        pass

    @abstractmethod
    def prgm_resume(self):
        '''
        Resume the paused program.
        '''
        pass

    @abstractmethod
    def prgm_next_step(self):
        '''
        Skip to the next step of the running program.
        '''
        pass

    @abstractmethod
    def get_prgm_counter(self):
        '''
        Get the status of the jump step/counter

        Returns:
            [{'name':str, 'start':int, 'end':int, 'cycles':int, 'remaining':int}]
        '''
        pass

    @abstractmethod
    def get_prgm_cur(self):
        '''
        Get the number of the running program

        Returns:
            int: The running program
        '''
        pass

    @abstractmethod
    def get_prgm_cstep(self):
        '''
        Get the current step of the running program.

        Returns:
            int: The current step of the running program
        '''
        pass

    @abstractmethod
    def get_prgm_cstime(self):
        '''
        Get the remaining time of the current step of the running program

        Returns:
            str: HH:MM:SS
        '''
        pass

    @abstractmethod
    def get_prgm_time(self, pgm=None):
        '''
        Get the time until the running program ends.

        Args:
            pgm (dict): The current program cache (optional,speeds up some controllers)
        Returns:
            str: HH:MM
        '''
        pass

    @abstractmethod
    def get_prgm_name(self, N):
        '''
        Get the name of a given program.

        Args:
            N (int): The program to get the name of
        Returns:
            str: The program name
        '''
        pass

    @abstractmethod
    def set_prgm_name(self, N, value):
        '''
        Set the name of a givem program

        Args:
            N (int): The program to set the name of
            value (str): The new name for the program
        '''
        pass

    @abstractmethod
    def get_prgm_steps(self, N):
        '''
        Get the total number of steps in a program

        Args:
            N (int): The program number
        Returns:
            int: The total number of steps
        '''
        pass

    @abstractmethod
    def get_prgms(self):
        '''
        Get a list of all programs on the chamber.

        Returns:
            [{"number":int, "name":str}]
        '''
        pass

    @abstractmethod
    def get_prgm(self, N):
        '''
        Get a program

        Args:
            N (int): The program number
        Returns:
            dict (format varies from controller to controller)
        '''
        pass

    @abstractmethod
    def set_prgm(self, N, value):
        '''
        Set a program

        Args:
            N (int): The program number
            value (dict): The program to write to the controller
        '''
        pass

    @abstractmethod
    def prgm_delete(self, N):
        '''
        Delete a program

        Args:
            N (int): The program number
        '''
        pass

    @exclusive
    def sample(self, lookup=None, **kwargs):
        '''
        Take a sample for data logging, gets datetime, mode, and sp/pv on all loops

        kwargs:
            get_loops (bool): If true get loop status (Default=True)
            get_status (bool): If true get controller operation status. (Default=True)
            get_alarms (bool): If true get the controller alarms status. (Default=False)
            get_program_status (bool): If true get information about the running progam. (Default=False)
            get_events (list or dict): A list of events to get the current status of (Default=None)
            get_program_list (bool): get a list of programs on the controller. (Default=False)
            get_refrig (bool): If true get the controller refrig mode. (Default=False)
        Returns:
            {"datetime":datetime.datetime, "loops":[{varies}], "status":str}
        '''
        ret = {'datetime':self.get_datetime(exclusive=False)}
        if kwargs.get('get_loops', True):
            ret['loops'] = []
            for tmap in self.loop_map:
                items = ['setpoint', 'processvalue', 'enable', 'mode', 'power']
                if tmap['type'] == 'cascade':
                    items.append('enable_cascade')
                lkps = [lkp for lkp in lookup[tmap['type']] if lkp['number'] == tmap['num']]
                lpdata = lkps[0].copy() if lookup else {}
                lpdata.update(self.get_loop(tmap['num'], tmap['type'], items, exclusive=False))
                ret['loops'].append(lpdata)
        if kwargs.get('get_status', True):
            ret['status'] = self.get_status(exclusive=False)
        if kwargs.get('get_alarms', False):
            ret['alarms'] = self.get_alarm_status(exclusive=False)
        if kwargs.get('get_program_status', False):
            if ret['status'].startswith('Program'):
                cpn = self.get_prgm_cur(exclusive=False)
                ret['program_status'] = {
                    'number':cpn,
                    'name':self.get_prgm_name(cpn, exclusive=False),
                    'step':self.get_prgm_cstep(exclusive=False),
                    'time_step':self.get_prgm_cstime(exclusive=False),
                    'time_total':self.get_prgm_time(kwargs.get('running_program'), exclusive=False)
                }
            else:
                ret['program_status'] = {'number':None, 'name':None, 'step':None}
        if kwargs.get('get_program_list', False):
            ret['program_list'] = self.get_prgms(exclusive=False)
        if kwargs.get('get_events', None):
            if isinstance(kwargs['get_events'][0], dict):
                events = kwargs['get_events']
            else:
                events = [{'N':i} for i in kwargs['get_events']]
            for event in events:
                event['status'] = self.get_event(event['N'], exclusive=False)
            ret['events'] = events
        if kwargs.get('get_refrig', False):
            try:
                ret['refrig'] = self.get_refrig(exclusive=False)
            except NotImplementedError:
                ret['refrig'] = None
        return ret

    @abstractmethod
    def process_controller(self, update=True):
        '''
        Read the controllers "part number" and setup the class as best as possible using it.

        Args:
            update (bool): When true update the classes configuration (default=True)
        Returns:
            str: The "part number"
        '''
        pass

    @abstractmethod
    def get_network_settings(self):
        '''
        Get the network settings from the controller (if controller supports)

        Returns:
            {"address":str, "mask":str, "gateway":str, "message":str, "host":str}
        '''
        pass

    @abstractmethod
    def set_network_settings(self, value):
        '''
        Set the network settings of the controller (if controller supports)

        Args:
            value ({"address":str, "mask":str, "gateway":str, "message":str, "host":str}): Settings
        '''
        pass

    def get_operation_modes(self):
        '''
        Get the supported operation modes for this controller.

        Returns:
            ["standby","constant","program"] or ["constant","program"]
        '''
        return ['standby', 'constant', 'program']

    def self_test(self, loops, cascades):
        '''
        preform a self test on all functions

        Args:
            loops (int): The number of standard control loops
            cascades (int): The number of cascade control loops
        '''

        def print_exception(trce):
            '''
            Format an Exception for printing
            '''
            print '\n'.join(['\t' + l for l in trce.split('\n')])

        print 'process_controller():'
        try:
            print '\t%r' % self.process_controller()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_datetime:'
        try:
            print '\t%r' % self.get_datetime()
        except Exception:
            print_exception(traceback.format_exc())

        print 'set_datetime:'
        try:
            self.set_datetime(self.get_datetime())
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_operation:'
        try:
            print '\t%r' % self.get_operation()
        except Exception:
            print_exception(traceback.format_exc())

        for lpi in range(loops):
            i = lpi + 1
            print 'get_loop_sp(%d):' % i
            try:
                print '\t%r' % self.get_loop_sp(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_loop_sp(%d):' %i
            try:
                self.set_loop_sp(i, self.get_loop_sp(i)['constant'])
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_loop_pv(%d):' % i
            try:
                print '\t%r' % self.get_loop_pv(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_loop_range(%d):' % i
            try:
                print '\t%r' % self.get_loop_range(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_loop_range(%d):' %i
            try:
                self.set_loop_range(i, self.get_loop_range(i))
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_loop_en(%d):' % i
            try:
                print '\t%r' % self.get_loop_en(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_loop_en(%d):' %i
            try:
                self.set_loop_en(i, self.get_loop_en(i)['constant'])
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_loop_units(%d):' % i
            try:
                print '\t%r' % self.get_loop_units(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_loop_mode(%d):' % i
            try:
                print '\t%r' % self.get_loop_mode(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_loop_power(%d):' % i
            try:
                print '\t%r' % self.get_loop_power(i)
            except Exception:
                print_exception(traceback.format_exc())

        for csi in range(cascades):
            i = csi + 1
            print 'get_cascade_sp[%d]:' % i
            try:
                print '\t%r' % self.get_cascade_sp(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_cascade_sp(%d):' %i
            try:
                self.set_cascade_sp(i, self.get_cascade_sp(i)['constant'])
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_pv(%d):' % i
            try:
                print '\t%r' % self.get_cascade_pv(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_range(%d):' % i
            try:
                print '\t%r' % self.get_cascade_range(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_cascade_range[%d]:' %i
            try:
                self.set_cascade_range(i, self.get_cascade_range(i))
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_en[%d]:' % i
            try:
                print '\t%r' % self.get_cascade_en(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_cascade_en(%d):' %i
            try:
                self.set_cascade_en(i, self.get_cascade_en(i)['constant'])
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_units(%d):' % i
            try:
                print '\t%r' % self.get_cascade_units(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_mode(%d):' % i
            try:
                print '\t%r' % self.get_cascade_mode(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_ctl(%d):' % i
            try:
                print '\t%r' % self.get_cascade_ctl(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_cascade_ctl(%d):' %i
            try:
                self.set_cascade_ctl(i, self.get_cascade_ctl(i))
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_cascade_deviation(%d):' % i
            try:
                print '\t%r' % self.get_cascade_deviation(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_cascade_deviation(%d):' %i
            try:
                self.set_cascade_deviation(i, self.get_cascade_deviation(i))
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

        for i in range(1, 13):
            print 'get_event(%d):' % i
            try:
                print '\t%r' % self.get_event(i)
            except Exception:
                print_exception(traceback.format_exc())
            print 'set_event(%d):' %i
            try:
                self.set_event(i, self.get_event(i)['current'])
                print '\tok'
            except Exception:
                print_exception(traceback.format_exc())

        print 'get_status:'
        try:
            print '\t%r' % self.get_status()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_alarm_status:'
        try:
            print '\t%r' % self.get_alarm_status()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_prgm_cur:'
        try:
            print '\t%r' % self.get_prgm_cur()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_prgm_cstep:'
        try:
            print '\t%r' % self.get_prgm_cstep()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_prgm_cstime:'
        try:
            print '\t%r' % self.get_prgm_cstime()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_prgm_time:'
        try:
            print '\t%r' % self.get_prgm_time()
        except Exception:
            print_exception(traceback.format_exc())

        for i in range(1, 6): #do 5 programs only
            print 'get_prgm_name(%d):' % i
            try:
                print '\t%r' % self.get_prgm_name(i)
            except Exception:
                print_exception(traceback.format_exc())

            print 'get_prgm_steps(%d):' % i
            try:
                print '\t%r' % self.get_prgm_steps(i)
            except Exception:
                print_exception(traceback.format_exc())

        print 'get_prgms:'
        try:
            print '\t%r' % self.get_prgms()
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_prgm(1):'
        try:
            print '\t%r' % self.get_prgm(1)
        except Exception:
            print_exception(traceback.format_exc())
        print 'set_prgm(1):'
        try:
            self.set_prgm(2, self.get_prgm(1))
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'get_network_settings:'
        try:
            print '\t%r' % self.get_network_settings()
        except Exception:
            print_exception(traceback.format_exc())
        print 'set_network_settings:'
        try:
            self.set_network_settings(self.get_network_settings())
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'call const_start():'
        try:
            self.const_start()
            time.sleep(5)
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'call stop():'
        try:
            self.stop()
            time.sleep(5)
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'call prgm_start(1,1):'
        try:
            self.prgm_start(1, 1)
            time.sleep(5)
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'call prgm_pause():'
        try:
            self.prgm_pause()
            time.sleep(5)
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'call prgm_resume():'
        try:
            self.prgm_resume()
            time.sleep(5)
            print '\tok'
        except Exception:
            print_exception(traceback.format_exc())

        print 'call sample():'
        try:
            print '\t%r' % self.sample()
        except Exception:
            print_exception(traceback.format_exc())

        print 'Testing Complete'



'''
Upper level interface for Espec Corp. Controllers (just the P300 for now)

:copyright: (C) Espec North America, INC.
:license: MIT, see LICENSE for more details.
'''
#pylint: disable=R0902,R0904
import datetime
import time
from chamberconnectlibrary.controllerinterface import ControllerInterface, exclusive
from chamberconnectlibrary.controllerinterface import ControllerInterfaceError
from chamberconnectlibrary.p300 import P300
from chamberconnectlibrary.scp220 import SCP220
from chamberconnectlibrary.especinteract import EspecError

class Espec(ControllerInterface):
    '''
    A class for interfacing with Espec controllers (P300, SCP220)

    Kwargs:
        interface (str): The connection method::
            "TCP" -- Use a Ethernet to serial adapter with raw TCP
            "Serial" -- Use a hardware serial port
        adr (int): The address of the controller (default=1)
        host (str): The hostname (IP address) of the controller when interface="TCP"
        serialport (str): The serial port to use when interface="Serial" (default=3(COM4))
        baudrate (int): The serial port's baud rate to use when interface="Serial" (default=9600)
        loops (int): The number of control loops the controller has (default=1, max=2)
        cascades (int): The number of cascade control loops the controller has (default=0, max=1)
        lock (RLock): The locking method to use when accessing the controller (default=RLock())
        freshness (int): The length of time (in seconds) a command is cached (default = 0)
        ctlr_type (str): "SCP220" or "P300" (default = "P300")
    '''

    def __init__(self, **kwargs):
        self.client, self.loops, self.cascades = None, None, None
        self.init_common(**kwargs)
        self.freshness = kwargs.get('freshness', 0)
        self.cache = {}
        self.temp, self.humi = 1, 2
        self.lpd = {
            'temp':self.temp,
            'humi':self.humi,
            'temperature':self.temp,
            'humidity':self.humi,
            'Temperature':self.temp,
            'Humidity':self.humi,
            self.temp:self.temp,
            self.humi:self.humi
        }
        self.ctlr_type = kwargs.get('ctlr_type', 'P300')
        ttp = (self.ctlr_type, self.temp, self.humi)
        self.lp_exmsg = 'The %s controller only supports 2 loops (%d:temperature,%d:humidity)'%ttp
        ttp = (self.ctlr_type, self.temp)
        self.cs_exmsg = 'The %s controller can only have loop %d as cascade' % ttp
        self.alarms = 27
        self.profiles = True
        self.events = 12
        self.total_programs = 40 if self.ctlr_type == 'P300' else 30
        self.__update_loop_map()


    def __update_loop_map(self):
        '''
        update the loop map.
        '''
        if self.cascades > 0:
            self.loop_map = [{'type':'cascade', 'num':1}]
        else:
            self.loop_map = [{'type':'loop', 'num':1}]
        if self.cascades + self.loops > 1:
            self.loop_map += [{'type':'loop', 'num':2}]
        self.named_loop_map = {'Temperature':0, 'temperature':0, 'Temp':0, 'temp':0}
        if len(self.loop_map) > 1:
            self.named_loop_map.update({'Humidity':1, 'humidity':1, 'Hum':1, 'hum':1})

    def connect(self):
        '''
        connect to the controller using the paramters provided on class initialization
        '''
        args = {'serialport':self.serialport, 'baudrate':self.baudrate, 'host':self.host,
                'address':self.adr}
        if self.ctlr_type == 'P300':
            self.client = P300(self.interface, **args)
        elif self.ctlr_type == 'SCP220':
            self.client = SCP220(self.interface, **args)
        else:
            raise ValueError('"%s" is not a supported controller type' % self.ctlr_type)

    def close(self):
        '''
        close the connection to the controller
        '''
        try:
            self.client.close()
        except AttributeError:
            pass
        self.client = None

    def cached(self, func, *args, **kwargs):
        '''
        The P300 returns multiple parameters with each command. The commands responses will be
        cached and cached responses returned if they are fresh enough (settable property)
        '''
        now = time.time()
        incache = func.__name__ not in self.cache
        if incache or (now - self.cache[func.__name__]['timestamp'] > self.freshness):
            self.cache[func.__name__] = {'timestamp':now, 'values':func(*args, **kwargs)}
        return self.cache[func.__name__]['values']

    @exclusive
    def raw(self, command):
        '''
        connect directly to the controller
        '''
        try:
            return self.client.interact(command)
        except EspecError as exc:
            emsg = str(exc)
            if 'The chamber did not respond in time' in emsg:
                return 'NA: SERIAL TIMEOUT'
            qps = [i for i, c in enumerate(emsg) if c == '"']
            return 'NA:' + emsg[qps[len(qps)-2]+1:qps[len(qps)-1]]

    @exclusive
    def get_refrig(self):
        return self.client.read_constant_ref()

    @exclusive
    def set_refrig(self, value):
        self.client.write_set(**value)

    @exclusive
    def set_loop(self, identifier, loop_type='loop', param_list=None, **kwargs):
        #cannot use the default controllerInterface version.
        lpfuncs = {
            'cascade':{
                'setpoint':self.set_cascade_sp,
                'setPoint':self.set_cascade_sp,
                'setValue':self.set_cascade_sp,
                'range':self.set_cascade_range,
                'enable':self.set_cascade_en,
                'deviation':self.set_cascade_deviation,
                'enable_cascade':self.set_cascade_ctl,
                'mode': self.set_cascade_mode},
            'loop':{
                'setpoint':self.set_loop_sp,
                'setPoint':self.set_loop_sp,
                'range':self.set_loop_range,
                'enable':self.set_loop_en,
                'mode':self.set_loop_mode
            }
        }
        if param_list is None:
            param_list = kwargs
        if isinstance(identifier, basestring):
            my_loop_map = self.loop_map[self.named_loop_map[identifier]]
            loop_number = my_loop_map['num']
            loop_type = my_loop_map['type']
        elif isinstance(identifier, (int, long)):
            loop_number = identifier
        else:
            raise ValueError(
                'invalid argument format, call w/: '
                'set_loop(int(identifier), str(loop_type), **kwargs) or '
                'get_loop(str(identifier), **kwargs)'
            )
        spt1 = 'setpoint' in param_list
        spt2 = 'setPoint' in param_list
        spt3 = 'setValue' in param_list
        if (spt1 or spt2 or spt3) and ('enable' in param_list or 'mode' in param_list):
            if 'enable' in param_list:
                enable = param_list.pop('enable')
                if isinstance(enable, dict):
                    enable = enable['constant']
            else:
                my_mode = param_list.pop('mode')
                if isinstance(my_mode, dict):
                    my_mode = my_mode['constant']
                enable = my_mode in ['On', 'ON', 'on']
            if spt1:
                spv = param_list.pop('setpoint')
            elif spt2:
                spv = param_list.pop('setPoint')
            else:
                spv = param_list.pop('setValue')
            if isinstance(spv, dict):
                spv = spv['constant']
            params = {'setpoint':spv, 'enable':enable}
            if range in param_list:
                params.update(param_list.pop('range'))
            if self.lpd[loop_number] == self.temp:
                self.client.write_temp(**params)
            elif self.lpd[loop_number] == self.humi:
                self.client.write_humi(**params)
            else:
                raise ValueError(self.lp_exmsg)
        if 'deviation' in param_list and 'enable_cascade' in param_list:
            if isinstance(param_list['enable_cascade'], dict):
                params = {'enable':param_list.pop('enable_cascade')['constant']}
            else:
                params = {param_list.pop('enable_cascade')}
            params.update(param_list.pop('deviation'))
            self.client.write_temp_ptc(**params)
        for key, val in param_list.items():
            params = {'value':val}
            params.update({'exclusive':False, 'N':loop_number})
            try:
                lpfuncs[loop_type][key](**params)
            except KeyError:
                pass

    @exclusive
    def get_datetime(self):
        temp = self.client.read_time()
        temp.update(self.client.read_date())
        return datetime.datetime(**temp)

    @exclusive
    def set_datetime(self, value):
        weekday = ['MON', 'TUE', 'WED', 'THU', 'FRI', 'SAT', 'SUN'][value.weekday()]
        self.client.write_time(value.hour, value.minute, value.second)
        self.client.write_date(value.year, value.month, value.day, weekday)

    @exclusive
    def get_loop_sp(self, N):
        if N not in self.lpd:
            raise ValueError(self.lp_exmsg)
        if self.lpd[N] == self.temp:
            cur = self.cached(self.client.read_temp)['setpoint']
            con = self.cached(self.client.read_constant_temp)['setpoint']
        else:
            cur = self.cached(self.client.read_humi)['setpoint']
            con = self.cached(self.client.read_constant_humi)['setpoint']
        return {'constant':con, 'current':cur}

    @exclusive
    def set_loop_sp(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if self.lpd[N] == self.temp:
            self.client.write_temp(setpoint=value)
        elif self.lpd[N] == self.humi:
            self.client.write_humi(setpoint=value)
        else:
            raise ValueError(self.lp_exmsg)

    @exclusive
    def get_loop_pv(self, N):
        if self.lpd[N] == self.temp:
            return {'air':self.cached(self.client.read_temp)['processvalue']}
        elif self.lpd[N] == self.humi:
            return {'air':self.cached(self.client.read_humi)['processvalue']}
        else:
            raise ValueError(self.lp_exmsg)

    @exclusive
    def set_loop_range(self, N, value):
        if 'max' not in value or 'min' not in value:
            raise AttributeError('missing "max" or "min" property')
        if self.lpd[N] == self.temp:
            self.client.write_temp(min=value['min'], max=value['max'])
        elif self.lpd[N] == self.humi:
            self.client.write_humi(min=value['min'], max=value['max'])
        else: raise ValueError(self.lp_exmsg)

    @exclusive
    def get_loop_range(self, N):
        if self.lpd[N] == self.temp:
            return self.cached(self.client.read_temp)['range']
        elif self.lpd[N] == self.humi:
            return self.cached(self.client.read_humi)['range']
        else: raise ValueError(self.lp_exmsg)

    @exclusive
    def get_loop_en(self, N):
        if self.lpd[N] == self.temp:
            return {'constant':True, 'current':True}
        elif self.lpd[N] == self.humi:
            return {'current':self.cached(self.client.read_humi)['enable'],
                    'constant':self.cached(self.client.read_constant_humi)['enable']}
        else: raise ValueError(self.lp_exmsg)

    @exclusive
    def set_loop_en(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if self.lpd[N] == self.temp:
            pass
        elif self.lpd[N] == self.humi:
            if value:
                self.client.write_humi(
                    enable=True,
                    setpoint=self.cached(self.client.read_constant_humi)['setpoint']
                )
            else:
                self.client.write_humi(enable=False)
        else: raise ValueError(self.lp_exmsg)

    @exclusive
    def get_loop_units(self, N):
        if self.lpd[N] == self.temp:
            return u'\xb0C'
        elif self.lpd[N] == self.humi:
            return u'%RH'
        else:
            raise ValueError(self.lp_exmsg)

    @exclusive
    def set_loop_mode(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if N > 2:
            raise ValueError(self.lp_exmsg)
        if value in ['Off', 'OFF', 'off']:
            self.set_loop_en(N, False, exclusive=False)
        elif value in ['On', 'ON', 'on']:
            self.set_loop_en(N, True, exclusive=False)
        else:
            raise ValueError('Mode must be on or off, recived:' + value)

    @exclusive
    def get_loop_mode(self, N):
        if N > 2:
            raise ValueError(self.lp_exmsg)
        if self.lpd[N] == self.humi:
            cur = 'On' if self.cached(self.client.read_humi)['enable'] else 'Off'
            con = 'On' if self.cached(self.client.read_constant_humi)['enable'] else 'Off'
        else:
            cur = 'On'
            con = 'On'
        if self.client.read_mode() in ['OFF', 'STANDBY']:
            cur = 'Off'
        return {"current": cur, "constant": con}

    def get_loop_modes(self, N):
        if N == 1:
            return ['On']
        elif N == 2:
            return ['Off', 'On']
        else:
            raise ValueError(self.lp_exmsg)

    @exclusive
    def get_loop_power(self, N):
        if self.lpd[N] == self.temp:
            val = self.cached(self.client.read_htr)['dry']
        elif self.lpd[N] == self.humi:
            val = self.cached(self.client.read_htr)['wet']
        else:
            raise ValueError(self.lp_exmsg)
        return {'current':val, 'constant':val}

    def set_loop_power(self, N, value):
        raise NotImplementedError

    @exclusive
    def get_cascade_sp(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        cur = self.cached(self.client.read_temp_ptc)
        enc = cur['enable_cascade']
        return {
            'constant':self.cached(self.client.read_constant_temp)['setpoint'],
            'current':cur['setpoint']['product'] if enc else cur['setpoint']['air'],
            'air':cur['setpoint']['air'],
            'product':cur['setpoint']['product']
        }

    @exclusive
    def set_cascade_sp(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        self.client.write_temp(setpoint=value)

    @exclusive
    def get_cascade_pv(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.cached(self.client.read_temp_ptc)['processvalue']

    @exclusive
    def get_cascade_range(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.get_loop_range(self.temp, exclusive=False)

    @exclusive
    def set_cascade_range(self, N, value):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        self.set_loop_range(self.temp, value, exclusive=False)

    @exclusive
    def get_cascade_en(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.get_loop_en(self.temp, exclusive=False)

    @exclusive
    def set_cascade_en(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.set_loop_en(self.temp, value, exclusive=False)

    @exclusive
    def get_cascade_units(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.get_loop_units(self.temp, exclusive=False)

    @exclusive
    def set_cascade_mode(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.set_loop_mode(N, value, exclusive=False)

    @exclusive
    def get_cascade_mode(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.get_loop_mode(self.temp, exclusive=False)

    def get_cascade_modes(self, N):
        return self.get_loop_modes(N)

    @exclusive
    def get_cascade_ctl(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return {
            'current': self.cached(self.client.read_temp_ptc)['enable_cascade'],
            'constant': self.cached(self.client.read_constant_ptc)['enable']
        }

    @exclusive
    def set_cascade_ctl(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        params = self.cached(self.client.read_temp_ptc)
        params['deviation'].update({'enable':value})
        self.client.write_temp_ptc(**params['deviation'])

    @exclusive
    def get_cascade_deviation(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.cached(self.client.read_constant_ptc)['deviation']

    @exclusive
    def set_cascade_deviation(self, N, value):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        if 'positive' not in value or 'negative' not in value:
            raise ValueError('value must contain "positive" and "negative" keys')
        self.client.write_temp_ptc(self.get_cascade_ctl(self.temp, exclusive=False), **value)

    @exclusive
    def get_cascade_power(self, N):
        if self.lpd[N] != self.temp:
            raise ValueError(self.cs_exmsg)
        return self.get_loop_power(self.temp, exclusive=False)

    @exclusive
    def set_cascade_power(self, N, value):
        raise NotImplementedError

    @exclusive
    def get_event(self, N):
        if N >= 13:
            raise ValueError('There are only 12 events')
        return {
            'current':self.cached(self.client.read_relay)[N-1],
            'constant':self.cached(self.client.read_constant_relay)[N-1]
        }

    @exclusive
    def set_event(self, N, value):
        value = value['constant'] if isinstance(value, dict) else value
        if N >= 13:
            raise ValueError('There are only 12 events')
        self.client.write_relay([value if i == N else None for i in range(1, 13)])

    @exclusive
    def get_status(self):
        if self.cached(self.client.read_mon)['alarms'] > 0:
            return 'Alarm'
        return {
            'OFF':'Off',
            'STANDBY':'Standby',
            'CONSTANT':'Constant',
            'RUN':'Program Running',
            'RUN PAUSE':'Program Paused',
            'RUN END HOLD':'Program End Hold',
            'RMT RUN':'Remote Program Running',
            'RMT RUN PAUSE':'Remote Program Paused',
            'RMT RUN END HOLD':'Remote Program End Hold'
        }[self.client.read_mode(True)]

    @exclusive
    def get_alarm_status(self):
        active = self.client.read_alarm()
        alarmlist = [0, 1, 2, 3, 6, 7, 8, 9, 10, 11, 12, 18, 19, 21, 22, 23, 26,
                     30, 31, 40, 41, 43, 46, 48, 50, 51, 99]
        inactive = [x for x in alarmlist if x not in active]
        return {'active':active, 'inactive':inactive}

    @exclusive
    def const_start(self):
        self.client.write_mode_constant()

    @exclusive
    def stop(self):
        self.client.write_mode_standby()

    @exclusive
    def prgm_start(self, N, step):
        self.client.write_prgm_run(N, step)

    @exclusive
    def prgm_pause(self):
        self.client.write_prgm_pause()

    @exclusive
    def prgm_resume(self):
        self.client.write_prgm_continue()

    @exclusive
    def prgm_next_step(self):
        self.client.write_prgm_advance()

    @exclusive
    def get_prgm_counter(self):
        prgm_set = self.client.read_prgm_set()
        prgm_data = self.client.read_prgm_data(prgm_set['number'])
        prgm_mon = self.client.read_prgm_mon()
        ret = [
            {'name':'A', 'remaining': prgm_mon['counter_a']},
            {'name':'B', 'remaining': prgm_mon['counter_b']},
        ]
        ret[0].update(prgm_data['counter_a'])
        ret[1].update(prgm_data['counter_b'])
        return ret

    @exclusive
    def get_prgm_cur(self):
        return self.cached(self.client.read_prgm_set)['number']

    @exclusive
    def get_prgm_cstep(self):
        return self.cached(self.client.read_prgm_mon)['pgmstep']

    @exclusive
    def get_prgm_cstime(self):
        rtime = self.cached(self.client.read_prgm_mon)['time']
        return '%d:%02d:00' % (rtime['hour'], rtime['minute'])

    @exclusive
    def get_prgm_time(self, pgm=None):
        if pgm is None:
            pgm = self.client.read_prgm(self.cached(self.client.read_prgm_set)['number'])
        pgms = self.cached(self.client.read_prgm_mon)

        #counter_a must be the inner counter or the only counter
        if (pgm['counter_a']['end'] >= pgm['counter_b']['end'] and \
           pgm['counter_a']['start'] <= pgm['counter_b']['start']) or \
           pgm['counter_a']['cycles'] == 0 and pgm['counter_b']['cycles'] != 0:
            pgm['counter_a'], pgm['counter_b'] = pgm['counter_b'], pgm['counter_a']
            pgms['counter_a'], pgms['counter_b'] = pgms['counter_b'], pgms['counter_a']
        cap = len(pgm['steps'])
        cap = cap*(pgm['counter_a']['cycles'] + 2) if pgm['counter_a']['cycles'] else cap
        cap = cap*(pgm['counter_b']['cycles'] + 2) if pgm['counter_b']['cycles'] else cap
        cap += 2

        cnta = pgms['counter_a']
        cntb = pgms['counter_b']
        cstp = pgms['pgmstep']-1
        tminutes = pgms['time']['hour']*60 + pgms['time']['minute']
        pcntb = cntb
        while cstp < len(pgm['steps']) and cap:
            cap -= 1
            if pgm['counter_a']['start'] == pgm['counter_b']['start'] and \
               pgm['counter_a']['cycles'] and pgm['counter_b']['cycles'] and \
               cstp == pgm['counter_b']['start']-1:
                if pcntb != cntb:
                    cnta = pgm['counter_a']['cycles']
            elif cstp == pgm['counter_b']['start']-1 and pgm['counter_b']['cycles']:
                cnta = pgm['counter_a']['cycles']
            if cstp == pgm['counter_a']['end']-1 and pgm['counter_a']['cycles'] and cnta:
                cstp = pgm['counter_a']['start']-1
                cnta -= 1
            elif cstp == pgm['counter_b']['end']-1 and pgm['counter_b']['cycles'] and cntb:
                cstp = pgm['counter_b']['start']-1
                pcntb = cntb
                cntb -= 1
            else:
                pcntb = cntb
                cstp += 1
            if cstp < len(pgm['steps']):
                tminutes += pgm['steps'][cstp]['time']['hour']*60 + \
                            pgm['steps'][cstp]['time']['minute']
        if cap == 0:
            raise RuntimeError('Calculating the total program time remaining aborted.')
        return "%d:%02d:00" % (int(tminutes/60), tminutes%60)

    @exclusive
    def get_prgm_name(self, N):
        return self.cached(self.client.read_prgm_data, N)['name']

    def set_prgm_name(self, N, value):
        raise NotImplementedError

    @exclusive
    def get_prgm_steps(self, N):
        return self.client.read_prgm_data(N)['steps']

    @exclusive
    def get_prgms(self):
        names = []
        for i in range(1, self.total_programs+1):
            try:
                names.append({'number':i, 'name':self.client.read_prgm_use_num(i)['name']})
            except EspecError:
                names.append({'number':i, 'name':''})
        return names

    @exclusive
    def get_prgm(self, N):
        try:
            return self.client.read_prgm(N, self.cascades > 0)
        except EspecError:
            raise ControllerInterfaceError('Could not read program from chamber controller.')

    @exclusive
    def set_prgm(self, N, prgm):
        self.client.write_prgm(N, prgm)

    @exclusive
    def prgm_delete(self, N):
        self.client.write_prgm_erase(N)


    @exclusive
    def process_controller(self, update=True):
        if update:
            self.loops = 0
            self.cascades = 0
        msg = 'P300 ' if self.client.read_rom().startswith('P3') else 'SCP-220 '
        try:
            if update:
                self.client.read_temp_ptc()
                self.cascades = 1
            msg += 'W/PTCON '
        except EspecError:
            if update:
                self.loops += 1
        try:
            if update:
                self.client.read_humi()
                self.loops += 1
            msg += 'W/Humidity'
        except EspecError:
            pass
        self.__update_loop_map()
        return msg

    @exclusive
    def get_network_settings(self):
        ret = self.client.read_ip_set()
        ret.update({'message':'', 'host':''})
        return ret

    @exclusive
    def set_network_settings(self, value):
        if value:
            self.client.write_ip_set(value.get('address', '0.0.0.0'),
                                     value.get('mask', '0.0.0.0'), value.get('gateway', '0.0.0.0'))
        else:
            self.client.write_ip_set('0.0.0.0', '0.0.0.0', '0.0.0.0')
